#tree {
    min-height: 400px;
    width: 100%;
    overflow: hidden;
}

#coverage-matrix_wrapper thead th {
    /* Turn headers 90° */
    writing-mode: vertical-rl;
    -webkit-transform: rotate(-180deg);
    -moz-transform: rotate(-180deg);
    min-height: 24px;
    min-width: 24px;
    padding: 0 0 0 0 !important
}

.toggle-vis {
    margin: 3px 0 3px 0;
}

table.dataTable tbody th, table.dataTable tbody td {
    padding: 0 0;
    text-align: center;
}

table, ::after, ::before {
    box-sizing: unset;
}

#coverage-matrix, #result-tsv {
    width: min-content !important;
    white-space: nowrap;
}

.plot-title {
    white-space: nowrap;
}

.plot-title h1, h2, h4 {
    display: inline;
}

@media screen and (min-width: 1100px) {
    #container-left {
        max-width: 65%;
        width: 65%;
        float: left;
        height: 100vh;
        overflow-x: auto;
        overflow-y: auto;
    }

    #container-right {
        max-width: 35%;
        width: 35%;
        float: left;
        height: 100vh;
        overflow-x: auto;
        overflow-y: auto;
    }

    #pie-container, #histogram-container {
        height: 50%;
        min-height: 300px;
    }

    #pie, #histogram {
        height: calc(100% - 50px);
        width: calc(100% - 10px);
        overflow: hidden;
    }
}

#result-tsv_wrapper, #coverage-matrix_wrapper {
    width: 100%;
    overflow-x: auto;
}

.plot-container {
    text-align: center;
    width: 100%;
}

.svg-container {
    display: inline-block;
}

.whitespace {
    white-space: pre;
    line-break: anywhere;
}

code {
    color: black;
}

.popover {
    width: min-content !important;
}

.popover-header {
    white-space: nowrap;
}

.main-svg {
    background: unset !important;
}

.title-with-link {
    display: flex;
    justify-content: space-between;
}

