#grid {
    display: grid;
    width: 100%;
    height: 100vh;
    grid-template-columns: 4fr 5fr;
}

.left {
    height: 100vh;
    min-width: 610px;
    max-width: 770px;
    overflow: scroll;
}

.right {
    height: 100vh;
    min-width: 400px;
    display: flex;
    flex-direction: column;
    justify-content: space-between;
    overflow-y: auto;
}

.btn {
    margin-top: 5px;
}

#trait-select {
    padding: 1rem;
    margin-bottom: 1rem;
}

#trait-info {
    margin-bottom: 1rem;
    overflow-y: scroll;
}

#trait-info h2 a {
    color: unset;
    text-decoration: unset;
    display: flex;
    justify-content: space-between;
    margin: 0;
    padding: 0;
}

#trait-info h2 button {
    margin: 0;
}

#svg-container {
    overflow-y: scroll;
    overflow-x: hidden;
}

.placeholder {
    opacity: unset;
    background-color: unset;
}

#legend {
    margin-bottom: 0.7rem;
}

#legend td, #legend th {
    text-align: center;
    padding: 0 3px 0 3px;
}

#legend th {
    font-weight: bold;
}

#legend-link {
    color: unset;
    text-decoration: unset;
}

#legend-link h3 {
    text-align: center;
}

.legend-title {
    display: flex;
}

.legend-title .btn {
    padding: 1px 8px 1px 8px;
    margin-left: 5px;
    margin-top: 2px;
}